local NEAR = 0.1
local FAR = 1000
local FOV = math.rad(60)
local DEFAULT_DISTANCE = 4
local DEFAULT_HEADING = 45
local DEGREES = math.pi / 180

local M =
{
    window_width = 0,
    window_height = 0,
    correct_distance = 2.0,
    aspect = 1.0,
    distance = DEFAULT_DISTANCE,
    heading = DEFAULT_HEADING,
    position = vmath.vector3(0, 0, 0),
    target = vmath.vector3(0, 0, 0),
    up = vmath.vector3(0, 1, 0),
    projection = vmath.matrix4(),
    view = vmath.matrix4(),
    frustum = vmath.matrix4(),
    inv_projectview = vmath.matrix4(),
}

local tmp_v = vmath.vector4(0, 0, 0, 1.0)
local nv = vmath.vector4(0, 0, -1, 1)
local fv = vmath.vector4(0, 0, 1, 1)

local function update_frustum(self)
    self.frustum = self.projection * self.view
    self.inv_projectview = vmath.inv(self.frustum)
end

local function update_heading(self)
    local heading_rad = self.heading * DEGREES
    local dx = math.cos(heading_rad)
    local dz = math.sin(heading_rad)

    self.position.x = self.target.x + (dx * (self.distance + self.correct_distance))
    self.position.y = self.target.y + (self.distance + self.correct_distance)
    self.position.z = self.target.z + (dz * (self.distance + self.correct_distance))
    self.view = vmath.matrix4_look_at(self.position, self.target, self.up)
end

local function update_window(self, window_width, window_height)
    self.window_width = window_width
    self.window_height = window_height

    local w = math.max(1, self.window_width)
    local h = math.max(1, self.window_height)

    if h > w then
        local dif_dist = h - w
        local proc = dif_dist / w
        self.correct_distance = DEFAULT_DISTANCE * proc
    else
        self.correct_distance = 0
    end

    update_heading(self)

    self.aspect = w / h
    self.projection = vmath.matrix4_perspective(FOV, self.aspect, NEAR, FAR)

    update_frustum(self)
end

local function normalize_heading(self)
    while self.heading > 360 do
        self.heading = self.heading - 360
    end
    while self.heading < 0 do
        self.heading = self.heading + 360
    end
end

function M:init(window_width, window_height)
    update_window(self, window_width, window_height)
end

function M:resize(window_width, window_height)
    update_window(self, window_width, window_height)
end

function M:set_heading(heading)
    self.heading = heading
    normalize_heading(self)
    update_heading(self)
    update_frustum(self)
end

function M:set_target(x, y, z)
    self.target.x = x
    self.target.y = y
    self.target.z = z
    update_heading(self)
    update_frustum(self)
end

function M:world_to_screen(wx, wy, wz)
    local viewport_x, viewport_y = 0, 0
    tmp_v.x, tmp_v.y, tmp_v.z, tmp_v.w = wx, wy, wz, 1.0
    tmp_v = self.frustum * tmp_v

    if tmp_v.w ~= 0 then
        local w = 1.0 / tmp_v.w
        tmp_v = tmp_v * w
    end

    tmp_v.x = (tmp_v.x * 0.5 + 0.5) * self.window_width + viewport_x
    tmp_v.y = (tmp_v.y * 0.5 + 0.5) * self.window_height + viewport_y

    return tmp_v.x, tmp_v.y, tmp_v.z
end

function M:get_ray_screen_to_world(x, y)
    local x1 = (2.0 * x / self.window_width) - 1.0
    local y1 = (2.0 * y / self.window_height) - 1.0

    nv.x, nv.y = x1, y1
    fv.x, fv.y = x1, y1

    local np = self.inv_projectview * nv
    local fp = self.inv_projectview * fv

    np = np * (1.0 / np.w)
    fp = fp * (1.0 / fp.w)

    return np.x, np.y, np.z, fp.x, fp.y, fp.z
end

function M:screen_to_world_plane(x, y, planeNormal, pointOnPlane)
    local npx, npy, npz, fpx, fpy, fpz = self:get_ray_screen_to_world(x, y)
    local np, fp = vmath.vector3(npx, npy, npz), vmath.vector3(fpx, fpy, fpz)
    local denom = vmath.dot(planeNormal, fp - np)

    if denom == 0 then return end

    local numer = vmath.dot(planeNormal, pointOnPlane - np)

    return vmath.lerp(numer / denom, np, fp)
end

return M
